@@ -31,6 +31,21 @@ struct gvm_configuration {
 	gboolean autoprinter;
 	char *autoprinter_command;
 	
+	gboolean autoscanner;
+	char *autoscanner_command;
+	
+	gboolean autokeyboard;
+	char *autokeyboard_command;
+	
+	gboolean automouse;
+	char *automouse_command;
+	
+	gboolean autotablet;
+	char *autotablet_command;
+	
+	gboolean autovideocam;
+	char *autovideocam_command;
+	
 	char *eject_command;
 };
 
