@@ -50,13 +50,11 @@
 static struct gvm_configuration config;
 static LibHalContext *hal_ctx;
 
-#ifndef ASSUME_SUBMOUNT
 /** Table of UDI's for volumes being mounted by g-v-m that we need to apply policy to */
 static GHashTable *mount_table = NULL;
 
 /** List of UDI's of all volumes mounted during the lifetime of the program */
 static GSList *mounted_volumes = NULL;
-#endif /* ! ASSUME_SUBMOUNT */
 
 
 typedef enum {
@@ -486,9 +484,9 @@ gvm_init_config (void)
  * gvm_run_command - run the given command, replacing %d with the device node,
  * %h with the HAL UDI and %m with the given path.
  *
- * Returns the exit status.
+ * Returns TRUE if successful or FALSE otherwise.
  */
-static int
+static gboolean
 gvm_run_command (const char *command, const char *udi, const char *device, const char *mount_point)
 {
 	char *path, *mp = NULL, *dev = NULL;
@@ -496,7 +494,6 @@ gvm_run_command (const char *command, const char *udi, const char *device, const
 	GError *error = NULL;
 	GString *exec;
 	char *argv[4];
-	int status;
 	
 	g_assert (udi != NULL);
 	
@@ -552,15 +549,16 @@ gvm_run_command (const char *command, const char *udi, const char *device, const
 	argv[2] = exec->str;
 	argv[3] = NULL;
 	
-	g_spawn_async (g_get_home_dir (), argv, NULL, 0, NULL, NULL, &status, &error);
-	if (error) {
+	if (!g_spawn_async (g_get_home_dir (), argv, NULL, 0, NULL, NULL, NULL, &error)) {
 		warn ("failed to exec %s: %s", exec->str, error->message);
+		g_string_free (exec, TRUE);
 		g_error_free (error);
+		return FALSE;
 	}
 	
 	g_string_free (exec, TRUE);
 	
-	return status;
+	return TRUE;
 }
 
 /*
@@ -609,17 +607,63 @@ gvm_check_dvd (const char *udi, const char *device, const char *mount_point)
  * Returns TRUE if the device is a camera or FALSE otherwise.
  */
 static gboolean
-gvm_udi_is_camera (const char *udi, gboolean check_libgphoto2)
+gvm_udi_is_camera (const char *udi)
 {
+	char *storage_device = NULL, *physical_device = NULL;
+	gboolean is_camera = FALSE;
+	DBusError error;
+	
+	dbus_error_init (&error);
+	if (!(storage_device = libhal_device_get_property_string (hal_ctx, udi, "block.storage_device", &error))) {
+		warn ("cannot get block.storage_device property: %s", error.message);
+		if (dbus_error_is_set (&error))
+			dbus_error_free (&error);
+		goto out;
+	}
+	
+	if (!(physical_device = libhal_device_get_property_string (hal_ctx, storage_device, "storage.physical_device", &error))) {
+		warn ("cannot get storage.physical_device property: %s", error.message);
+		if (dbus_error_is_set (&error))
+			dbus_error_free (&error);
+		goto out;
+	}
+	
+	if ((is_camera = libhal_device_query_capability (hal_ctx, physical_device, "camera", NULL)))
+		dbg ("Camera detected: %s\n", udi);
+	
+ out:
+	
+	libhal_free_string (storage_device);
+	libhal_free_string (physical_device);
+	
+	return is_camera;
+}
+
+/*
+ * gvm_udi_is_ptp_camera - checks if the udi is a PTP camera device
+ *
+ * Returns TRUE if the device is a PTP camera with libgphoto2 support or FALSE otherwise.
+ */
+static gboolean
+gvm_udi_is_ptp_camera (const char *udi)
+{
+	gboolean is_ptp_camera = FALSE;
+	char *access_method;
+	
 	if (!libhal_device_query_capability (hal_ctx, udi, "camera", NULL))
 		return FALSE;
 	
-	if (check_libgphoto2 && !libhal_device_get_property_bool (hal_ctx, udi, "camera.libgphoto2_support", NULL))
+	if (!(access_method = libhal_device_get_property_string (hal_ctx, udi, "camera.access_method", NULL)))
 		return FALSE;
 	
-	dbg ("Camera detected: %s\n", udi);
+	if (!strcmp (access_method, "ptp")) {
+		dbg ("PTP Camera detected: %s\n", udi);
+		is_ptp_camera = TRUE;
+	}
 	
-	return TRUE;
+	libhal_free_string (access_method);
+	
+	return is_ptp_camera;
 }
 
 /*
@@ -662,17 +706,6 @@ gvm_udi_is_ipod (const char *udi)
 	return is_ipod;
 }
 
-/*
- * gvm_udi_is_printer - checks if the udi is a printer
- *
- * Returns TRUE if the device is a printer or FALSE otherwise.
- */
-static gboolean
-gvm_udi_is_printer (const char *udi)
-{
-	return libhal_device_query_capability (hal_ctx, udi, "printer", NULL);
-}
-
 /*
  * gvm_run_camera - launch the camera application
  */
@@ -746,7 +779,7 @@ gvm_run_printer (const char *udi)
 {
 	DBusError error;
 	char *device;
-
+	
 	if (config.autoprinter_command == NULL)
 		return;
 	
@@ -851,7 +884,7 @@ gvm_device_mounted (const char *udi)
 	}
 	
 	/* this is where the magic happens */
-	if (config.autophoto && gvm_udi_is_camera (udi, FALSE)) {
+	if (config.autophoto && gvm_udi_is_camera (udi)) {
 		if (gvm_prompt (GVM_PROMPT_IMPORT_CAMERA) == GVM_RESPONSE_IMPORT_PHOTOS)
 			gvm_run_camera (udi, device, mount_point);
 	} else if (config.autoipod && gvm_udi_is_ipod (udi)) {
@@ -886,7 +919,28 @@ gvm_device_mounted (const char *udi)
 	libhal_free_string (device);
 }
 
-#ifndef ASSUME_SUBMOUNT
+static gboolean
+gvm_udi_is_subfs_mount (const char *udi)
+{
+	int subfs = FALSE;
+	char **callouts;
+	int i;
+	
+	if ((callouts = libhal_device_get_property_strlist (hal_ctx, udi, "info.callouts.add", NULL))) {
+		for (i = 0; callouts[i] != NULL; i++) {
+			if (!strcmp (callouts[i], "hald-block-subfs")) {
+				dbg ("subfs to handle mounting of %s; skipping\n", udi);
+				subfs = TRUE;
+				break;
+			}
+		}
+		
+		libhal_free_string_array (callouts);
+	}
+	
+	return subfs;
+}
+
 /*
  * gvm_device_mount - mount the given device.
  *
@@ -895,7 +949,20 @@ gvm_device_mounted (const char *udi)
 static gboolean
 gvm_device_mount (const char *udi, const char *device, const char *mount_point)
 {
-	return gvm_run_command (MOUNT_COMMAND, udi, device, mount_point) == 0;
+	char *key;
+	
+	key = g_strdup (udi);
+	g_hash_table_insert (mount_table, key, key);
+	
+	dbg ("mounting %s...\n", udi);
+	if (gvm_run_command (MOUNT_COMMAND, udi, device, mount_point))
+		return TRUE;
+	
+	dbg ("mount failed: %s\n", udi);
+	g_hash_table_remove (mount_table, key);
+	g_free (key);
+	
+	return FALSE;
 }
 
 /*
@@ -906,9 +973,8 @@ gvm_device_mount (const char *udi, const char *device, const char *mount_point)
 static gboolean
 gvm_device_unmount (const char *udi, const char *device, const char *mount_point)
 {
-	return gvm_run_command (UNMOUNT_COMMAND, udi, device, mount_point) == 0;
+	return gvm_run_command (UNMOUNT_COMMAND, udi, device, mount_point);
 }
-#endif /* ! ASSUME_SUBMOUNT */
 
 /*
  * gvm_run_cdplay - if so configured, execute the user-specified CD player on
@@ -930,7 +996,7 @@ gvm_run_cdplayer (const char *udi, const char *device, const char *mount_point)
 static void
 gvm_ask_mixed (const char *udi)
 {
-	char *key, *device = NULL, *mount_point = NULL;
+	char *device = NULL, *mount_point = NULL;
 	int action = GVM_RESPONSE_NONE;
 	DBusError error;
 	
@@ -954,14 +1020,10 @@ gvm_ask_mixed (const char *udi)
 	
 	switch (action) {
 	case GVM_RESPONSE_BROWSE:
-#ifndef ASSUME_SUBMOUNT
-		gvm_device_mount (udi, device, NULL);
-		
-		key = g_strdup (udi);
-		g_hash_table_insert (mount_table, key, key);
-#else
-		gvm_device_mounted (udi);
-#endif
+		if (gvm_udi_is_subfs_mount (udi))
+			gvm_device_mounted (udi);
+		else
+			gvm_device_mount (udi, device, NULL);
 		break;
 	case GVM_RESPONSE_PLAY:
 		gvm_run_cdplayer (udi, device, NULL);
@@ -1064,18 +1126,12 @@ gvm_cdrom_policy (const char *udi)
 	} else if (has_audio && has_data) {
 		gvm_ask_mixed (udi);
 	} else if (has_data) {
-#ifndef ASSUME_SUBMOUNT
 		if (config.automount_media) {
-			char *key;
-			
-			gvm_device_mount (udi, device, NULL);
-			
-			key = g_strdup (udi);
-			g_hash_table_insert (mount_table, key, key);
+			if (!gvm_udi_is_subfs_mount (udi))
+				gvm_device_mount (udi, device, NULL);
+			else
+				gvm_device_mounted (udi);
 		}
-#else
-		gvm_device_mounted (udi);
-#endif
 	} else if (is_blank) {
 		if (gvm_device_is_writer (drive_udi))
 			gvm_run_cdburner (udi, device, NULL);
@@ -1223,32 +1279,49 @@ hal_device_added (LibHalContext *ctx __attribute__((__unused__)),
 				goto out;
 		}
 		
-#ifndef ASSUME_SUBMOUNT
 		if (config.automount_drives && mountable) {
-			if (gvm_automount_enabled (udi)) {
-				char *key;
-				
-				gvm_device_mount (udi, device, NULL);
-				
-				key = g_strdup (udi);
-				g_hash_table_insert (mount_table, key, key);
+			if (!gvm_udi_is_subfs_mount (udi)) {
+				if (gvm_automount_enabled (udi)) {
+					gvm_device_mount (udi, device, NULL);
+				} else {
+					dbg ("storage.automount_enabled_hint set to false on %s, not mounting\n", udi);
+				}
 			} else {
-				dbg ("storage.automount_enabled_hint set to false on %s, not mounting\n", udi);
+				gvm_device_mounted (udi);
 			}
 		}
-#else
-		if (mountable)
-			gvm_device_mounted (udi);
-#endif
-	} else {
-		/* if the device is a PTP camera with libgphoto2 support, launch the user-specified application */
-		if (gvm_udi_is_camera (udi, TRUE)) {
-			if (config.autophoto && gvm_prompt (GVM_PROMPT_IMPORT_CAMERA) == GVM_RESPONSE_IMPORT_PHOTOS)
-				gvm_run_camera (udi, NULL, NULL);
-		} else if (gvm_udi_is_printer (udi)) {
-			if (config.autoprinter)
-				gvm_run_printer (udi);
+	} else if (libhal_device_query_capability (hal_ctx, udi, "input", NULL)) {
+		/* input device (keyboard, mouse, wacom tablet, etc...) */
+		const char *command = NULL;
+		int autoexec = FALSE;
+		char *device;
+		
+		if (libhal_device_query_capability (hal_ctx, udi, "input.keyboard", NULL)) {
+			autoexec = config.autokeyboard;
+			command = config.autokeyboard_command;
+		} else if (libhal_device_query_capability (hal_ctx, udi, "input.mouse", NULL)) {
+			autoexec = config.automouse;
+			command = config.automouse_command;
+		} else if (libhal_device_query_capability (hal_ctx, udi, "input.tablet", NULL)) {
+			autoexec = config.autotablet;
+			command = config.autotablet_command;
 		}
+		
+		if (autoexec && command) {
+			if ((device = libhal_device_get_property_string (hal_ctx, udi, "input.device", &error))) {
+				gvm_run_command (command, udi, device, NULL);
+				libhal_free_string (device);
+			} else {
+				warn ("cannot get input.device property: %s", error.message);
+			}
+		}
+	} else if (libhal_device_query_capability (hal_ctx, udi, "printer", NULL)) {
+		if (config.autoprinter)
+			gvm_run_printer (udi);
+	} else if (gvm_udi_is_ptp_camera (udi)) {
+		/* if the device is a PTP camera with libgphoto2 support, launch the user-specified application */
+		if (config.autophoto && gvm_prompt (GVM_PROMPT_IMPORT_CAMERA) == GVM_RESPONSE_IMPORT_PHOTOS)
+			gvm_run_camera (udi, NULL, NULL);
 	}
 	
  out:
@@ -1299,7 +1372,7 @@ hal_device_lost_capability (LibHalContext *ctx __attribute__((__unused__)),
 }
 
 /** Invoked when a property of a device in the Global Device List is
- *  changed, and we have we have subscribed to changes for that device.
+ *  changed, and we have subscribed to changes for that device.
  *
  *  @param  ctx                 LibHal context
  *  @param  udi                 Univerisal Device Id
@@ -1312,7 +1385,6 @@ hal_property_modified (LibHalContext *ctx __attribute__((__unused__)),
 		       dbus_bool_t is_removed __attribute__((__unused__)), 
 		       dbus_bool_t is_added __attribute__((__unused__)))
 {
-#ifndef ASSUME_SUBMOUNT
 	char *mounting_udi;
 	dbus_bool_t val;
 	GSList *l, *n;
@@ -1325,14 +1397,16 @@ hal_property_modified (LibHalContext *ctx __attribute__((__unused__)),
 	if (val == TRUE) {
 		dbg ("Mounted: %s\n", udi);
 		
-		/* add to list of all volumes mounted during lifetime */
-		mounted_volumes = g_slist_append (mounted_volumes, g_strdup (udi));
-		
 		if ((mounting_udi = g_hash_table_lookup (mount_table, udi))) {
 			g_hash_table_remove (mount_table, udi);
 			g_free (mounting_udi);
 			
+			/* add to list of all volumes mounted during lifetime */
+			mounted_volumes = g_slist_append (mounted_volumes, g_strdup (udi));
+			
 			gvm_device_mounted (udi);
+		} else {
+			dbg ("not in mount queue: %s\n", udi);
 		}
 	} else {
 		dbg ("Unmounted: %s\n", udi);
@@ -1348,7 +1422,6 @@ hal_property_modified (LibHalContext *ctx __attribute__((__unused__)),
 			}
 		}
 	}
-#endif /* ASSUME_SUBMOUNT */
 }
 
 /** Invoked when a device in the GDL emits a condition that cannot be
@@ -1457,7 +1530,6 @@ gvm_do_hal_init (void)
 	return ctx;
 }
 
-#ifndef ASSUME_SUBMOUNT
 /** Attempt to mount all volumes; should be called on startup.
  *
  *  @param  ctx                 LibHal context
@@ -1483,7 +1555,10 @@ mount_all (LibHalContext *ctx)
 	}
 	
 	for (i = 0; i < num_volumes; i++) {
-		udi = volumes [i];
+		udi = volumes[i];
+		
+		if (gvm_udi_is_subfs_mount (udi))
+			continue;
 		
 		/* don't attempt to mount already mounted volumes */
 		if (!libhal_device_property_exists (ctx, udi, "volume.is_mounted", NULL)
@@ -1561,7 +1636,6 @@ unmount_all (LibHalContext *ctx)
 		}
 	}
 }
-#endif /* ! ASSUME_SUBMOUNT */
 
 
 static int sigterm_unix_signal_pipe_fds[2];
@@ -1596,12 +1670,10 @@ sigterm_iochn_data (GIOChannel *source,
 		g_error_free (err);
 		goto out;
 	}
-
+	
 	dbg ("Received SIGTERM, initiating shutdown\n");
 	
-#ifndef ASSUME_SUBMOUNT
 	unmount_all (hal_ctx);
-#endif
 	
 	gtk_main_quit ();
 	
@@ -1615,9 +1687,7 @@ gvm_die (GnomeClient *client __attribute__((__unused__)),
 {
 	dbg ("Received 'die', initiating shutdown\n");
 	
-#ifndef ASSUME_SUBMOUNT
 	unmount_all (hal_ctx);
-#endif
 	
 	gtk_main_quit ();
 }
@@ -1668,9 +1738,7 @@ main (int argc, char *argv[])
 	
 	mount_table = g_hash_table_new (g_str_hash, g_str_equal);
 	
-#ifndef ASSUME_SUBMOUNT
 	mount_all (hal_ctx);
-#endif
 	
 	gtk_main ();
 	
